///////////////////////
//// Jeroen's Code ////
///////////////////////

// includes

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include "src/cudaCode.h"

#define real float


// includes, project
#include <cutil.h>


//#define TREECODE

/*

TODO
- Complete neighbourhood lijst
*- Laten werkend als aantal particles stijgt
*- Configuratie bestand (settings en print settings) 
*- Zorgen dat we zelf particle grote beheren aangezien dit verschilt per data set
  (indexmodifier code)


*/

//The variable that holds the number of particles in the system
int particles;


/// HOST variables
float4* h_ixdata;
float4* h_ivdata;
float3* h_oresadata;
float4* h_oresadotdata;
float2* h_oNdata;
float*  localMinDistanceList; //Used incase we make use of the kernel for smaller steps that adds on the cpu
int*   h_neighbourResults;
float*  h_h2;

float4*  h_xIn;
float3*  h_vIn;


/// DEVICE variables
float4* d_ix;
float4* d_iv;

float*  d_ipredictTime;
float*  d_ieps2;
int*    d_iindex;
int*    d_ihighest;

int*     d_neighbourResults;
float*   d_h2;
float4*  d_xPred;
float4*  d_vPred;
float3*  d_oforce;
float4*  d_oadoti;
float2*  d_oNdata; //Neighbourhood data, x is index, y is distance

float4*  d_xIn;
float3*  d_vIn;


int memSizeFloat4 = sizeof(float4) ;
int memSizeFloat3 = sizeof(float3);
int memSizeFloat2 = sizeof(float) ;
int memSizeFloat  = sizeof(float) ;
int memSizeInt    = sizeof(int) ;

__constant__ int    numberOfParticles;  //Number of particles in the data set
__constant__ int    extraFactor;        //Extra factor used by the small kernel
__constant__ int    indexModifier;      //Number of the loop we are in when we process large N in smaller chuncks

int localExtraFactor     = 0;
int localGridSize        = 0;
int usedNumberOfThreads  = 128;

//Configuration file settings
int smallKernelEndValue       = 8192;
int smallKernelBlocks         = 64;
int configNumberOfThreads     = 128;
int smallKernelDataSetMinSize = smallKernelBlocks*configNumberOfThreads;
int debugPrints               = 1;
int fiveSecondLimit           = 128;
int MAX_THREADS               = 64*1024; //64k default
int numberOfPipes             = DEFAULT_NPIPES;    

//Stats summary, usefull for explaining performance drops
int numberOfCalls = 0;
int blockSizeSum  = 0;

int zero = 0;

  
dim3  grid( 1, 1, 1);
dim3  threads( 1, 1, 1);

// includes, kernels
#include "src/kernels.cu"

//The real number N, unmodified
int realParticles = 0;
//The number of particles rounded by numberOfThreads, but before
//increased for memory size
int tempParticles = 0;
//Variable to store the increased memory size
//needed to store in the multiple partiale results.
int smallResultsParticles = 0;


void initCUDA(int blocks, int threads, int n)
{
  //Lets first check if we have a GPU, TODO own code incase of multigpu.
  CUT_DEVICE_INIT();

  particles     = n;
  realParticles = n;

  //Read in configuration file
  FILE   *inp = fopen("kirin.cfg","r");
  if(inp != NULL)
  {
    printf("Configuration file found, reading settings...\n");
    fscanf(inp,"%d %d %d %d %d %d %d",&numberOfPipes, &smallKernelEndValue, &smallKernelBlocks, 
                                      &configNumberOfThreads, &debugPrints, &fiveSecondLimit);
    fclose(inp);
  }
  else
  {
    printf("No configuration file found, using default settings \n");
  }
  
  //Calculate the minim size of the data set to be able to use small kernels
  smallKernelDataSetMinSize = smallKernelBlocks*configNumberOfThreads;

  printf("Settings used by Kirin: \n");
  printf("Threads per block: \t\t\t\t\t %d \n",                           configNumberOfThreads);
  printf("Kernel for small time-steps offset: \t\t\t %d \n",              smallKernelEndValue);
  printf("Kernel for small time-steps minumum number of blocks:\t %d \n", smallKernelBlocks);
  printf("Debug prints: \t\t\t\t\t\t %d \n",                              debugPrints);
  printf("Five second limit: \t\t\t\t\t %d \n",                           fiveSecondLimit*1024);
  printf("Number Of pipes : \t\t\t\t\t %d \n",                           numberOfPipes);

  #ifdef TREECODE
    printf("Version optimized for treecode\n");
  #else
    printf("Normal version of the library\n");
  #endif

  //Make particles a multiple of numberOfThreads so we dont get
  //memory errors within the GPU. This is cheaper then
  //doing ``if'' comparisons on the GPU.
  int numberOfThreads = configNumberOfThreads;
  if(particles % numberOfThreads != 0)
  {
    int tempResult = particles / numberOfThreads;
    tempResult++;
    particles = tempResult* numberOfThreads;
  }

  //Next section determines the Maximum ammount of blocks/threads we can start
  //before we reach the 5second limit. The fivesecond limit has to be specified
  //in the configuration file, as (N/1024). So 131072 has to be specified as 128.
  //Then we divide and determine the closest power of 2, which is used as Max threads value.
  //Only do this for particles > 1024 else we get 0 devisions
  if(particles > 1024)
  {
    int temp1     = (fiveSecondLimit*fiveSecondLimit) / (particles / 1024);
    temp1          = log((double)temp1) / log(2.0);
    MAX_THREADS    = pow(2.0,(double)temp1)*1024;
  }

  //If we split the work over multiple bundles we need more memory space
  //to store the results. This calculated how many positions
  //have to be allocated. 
  tempParticles         = particles;
  smallResultsParticles = particles;
  if(particles <= smallKernelBlocks*numberOfThreads)
    smallResultsParticles = smallKernelBlocks*numberOfThreads;


  //Memory sizes
  int numberInBuffer  = particles; 
  int sizeFloat4      = numberInBuffer*sizeof(float4);
  int sizeFloat3      = numberInBuffer*sizeof(float3);
  //int sizeFloat       = numberInBuffer*sizeof(float); 
  int sizeInt         = numberInBuffer*sizeof(int);

  memSizeFloat4 = sizeof(float4)* particles ;
  memSizeFloat3 = sizeof(float3)* particles;
  memSizeFloat2  = sizeof(float2) * particles;
  memSizeFloat  = sizeof(float) * particles;
  memSizeInt    = sizeof(int)* particles ;

  // allocate device memory for input data set
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_ix,   sizeFloat4));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_iv,   sizeFloat4));

  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_iindex, sizeInt));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_ihighest, sizeof(int)));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_ieps2, sizeof(float)));
  
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_xPred, sizeFloat4));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_vPred, sizeFloat4));

  // allocate device memory for the i-particles 
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_xIn, sizeof(float4)* numberOfPipes));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_vIn, sizeof(float3)* numberOfPipes));

  // allocate device memory for results    
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_oforce, smallResultsParticles*sizeof(float3)));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_oadoti, smallResultsParticles*sizeof(float4)));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_oNdata, smallResultsParticles*sizeof(float2)));

  //Allocate memory for neighbourhood list, memory size is related to the number of pipelines
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_neighbourResults, sizeof(int)*(numberOfPipes*255)));
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_h2, sizeof(float)*(numberOfPipes))); 
  h_h2                = (float*) malloc(memSizeFloat);
  h_neighbourResults  = (int*)   malloc(sizeof(int)*(numberOfPipes*255));

  //Allocate host memory for the results
  h_oresadata           = (float3*) calloc( smallResultsParticles, sizeof(float3));
  h_oresadotdata        = (float4*) calloc(smallResultsParticles, sizeof(float4));
  h_oNdata              = (float2*)    malloc(smallResultsParticles*sizeof(float2));
  localMinDistanceList  = (float*) malloc(smallResultsParticles*sizeof(float));

  //Allocate host memory for the local buffers
  //I particles;
  h_ixdata = (float4*) malloc( sizeFloat4);
  h_ivdata = (float4*) malloc( sizeFloat4);
  //I particles
  h_xIn = (float4*) malloc(sizeof(float4)*numberOfPipes);
  h_vIn = (float3*) malloc(sizeof(float3)*numberOfPipes);

  
  //Set constants
  CUDA_SAFE_CALL(cudaMemcpyToSymbol(numberOfParticles,&tempParticles,sizeof(int),0));
 
  //Fill the force data memory with zeros since we will use
  //the memory before we set forces during the first prediction
  //TODO: Prediction on the device is not available  atm
  CUDA_SAFE_CALL( cudaMemcpy( d_oforce, h_oresadata, sizeFloat3,
                            cudaMemcpyHostToDevice) );
  CUDA_SAFE_CALL( cudaMemcpy( d_oadoti, h_oresadotdata, sizeFloat3,
                            cudaMemcpyHostToDevice) );

  // set texture parameters
  texX.addressMode[0] = cudaAddressModeWrap;
  texX.addressMode[1] = cudaAddressModeWrap;
  texX.filterMode     = cudaFilterModeLinear;
  texX.normalized     = false;    // access with normal coordinates, so [0..n]

  // set texture parameters
  texV.addressMode[0] = cudaAddressModeWrap;
  texV.addressMode[1] = cudaAddressModeWrap;
  texV.filterMode     = cudaFilterModeLinear;
  texV.normalized     = false;    // access with normal coordinates , so [0..n]

  // Bind the arrays to the textures
  CUDA_SAFE_CALL( cudaBindTexture(0, texX, d_xPred,sizeFloat4));
  CUDA_SAFE_CALL( cudaBindTexture(0, texV, d_vPred,sizeFloat4));
}


void sendToCUDAGPU(int *index,double3 *x, double3 *v, double *mass, int numberInBuffer)
{ 
  int sizeFloat4  = particles*sizeof(float4);
 
  //Convert the double data to floats
  for(int i=0; i < numberInBuffer; i++)
  {
    h_ixdata[i].x = (float)x[i].x;
    h_ixdata[i].y = (float)x[i].y;
    h_ixdata[i].z = (float)x[i].z;
    h_ixdata[i].w = (float)mass[i]; //mass is 4th component of position variable

    h_ivdata[i].x = (float)v[i].x;
    h_ivdata[i].y = (float)v[i].y;
    h_ivdata[i].z = (float)v[i].z;
 
   // h_iidata[i]  =  index[i];
  }

  //Fill the unused particle memory with zeros, so
  //they do not influence the end results when 
  //the values are used to get to a multiple of numberofthreads
  for(int i= numberInBuffer; i < particles; i++)
  {
    h_ixdata[i].x = 0;    h_ixdata[i].y = 0;
    h_ixdata[i].z = 0;    h_ixdata[i].w = 0;

    h_ivdata[i].x = 0;    h_ivdata[i].y = 0;
    h_ivdata[i].z = 0;    h_ivdata[i].w = 0;
  }


  // copy host memory to device
  //Position, velocity and mass
  CUDA_SAFE_CALL( cudaMemcpy( d_xPred, h_ixdata, sizeFloat4,
                              cudaMemcpyHostToDevice) );
  CUDA_SAFE_CALL( cudaMemcpy( d_vPred, h_ivdata, sizeFloat4,
                              cudaMemcpyHostToDevice) );

  //Set the indexModifier back to zero
  CUDA_SAFE_CALL(cudaMemcpyToSymbol(indexModifier,&zero,sizeof(int),0));
}

void calculateCUDAForces(float eps2, int *index,int nj, int blockSize, double xi[][3], double vi[][3], double h2[])
{
  //Some statistics sum
  blockSizeSum += blockSize;
  numberOfCalls++;

  /// Copy the i-particles to the GPU ///
  /// First we have to convert the doubles into floats and combine the various 
  /// tables.
  for(int i=0; i < blockSize; i++)
  {
    h_xIn[i].x = xi[i][0];    h_xIn[i].y = xi[i][1];
    h_xIn[i].z = xi[i][2];    h_xIn[i].w = h2[i];

    h_vIn[i].x = vi[i][0];    h_vIn[i].y = vi[i][1];
    h_vIn[i].z = vi[i][2];
  }


  //Copy the x values to device
  CUDA_SAFE_CALL( cudaMemcpy( d_xIn, h_xIn, blockSize*sizeof(float4),
                              cudaMemcpyHostToDevice) );
  //Copy the v values to device
  CUDA_SAFE_CALL( cudaMemcpy( d_vIn, h_vIn, blockSize*sizeof(float3),
                              cudaMemcpyHostToDevice) );


  //Shared memory size that we need
  int mem_size =2*sizeof(float4)*configNumberOfThreads ;

  //Incase of binaries the total number of particles is decreased
  //therefor change the settings.
  
  //TODO is dit nu wel of niet nodig....
  /*if(nj < particles)
  {
    //First set a new number for the particles number
   // CUDA_SAFE_CALL(cudaMemcpyToSymbol(numberOfParticles,&nj,sizeof(int),0));
    //Now put old data to 0's
    for(int i= nj; i < particles; i++)
    {
      h_ixdata[i].x = 0;      h_ixdata[i].y = 0;
      h_ixdata[i].z = 0;      h_ixdata[i].w = 0;

      h_ivdata[i].x = 0;      h_ivdata[i].y = 0;
      h_ivdata[i].z = 0;      h_ivdata[i].w = 0;
    }
   // particles = nj;

    CUDA_SAFE_CALL( cudaMemcpy( d_xPred, h_ixdata, particles*sizeof(float4),
                              cudaMemcpyHostToDevice) );
    CUDA_SAFE_CALL( cudaMemcpy( d_vPred, h_ivdata, particles*sizeof(float4),
                              cudaMemcpyHostToDevice) );
 }*/


    //Copy the eps value to device
  CUDA_SAFE_CALL( cudaMemcpy( d_ieps2, &eps2, sizeof(float),
                              cudaMemcpyHostToDevice) );

  if ((blockSize ) > MAX_THREADS)
  {
    //TODO....dit aanpassen
    //Start work around code
    //Calculate the number of loops
    int numberOfLoops = blockSize / MAX_THREADS;
    if(blockSize > (MAX_THREADS*numberOfLoops))
      numberOfLoops++;
    
    //For each block we have to do start the kernel
    for(int i=0; i < numberOfLoops; i++)
    {
      //Copy the increased value to the GPU
      int localIndexModifier = MAX_THREADS*i;
      CUDA_SAFE_CALL(cudaMemcpyToSymbol(indexModifier,&localIndexModifier,sizeof(int),0));

      printf("Nu in work around code indexm %d HU %d threads %d \n", localIndexModifier, blockSize,MAX_THREADS);
    
      //Calculate the number of grids that are needed
      threads.x = configNumberOfThreads;
      grid.x = (int)MAX_THREADS/threads.x;
      if(MAX_THREADS % threads.x != 0)
        grid.x = grid.x+1;

      //For the last step see if we can use less then MAX_THREADS threads
      if(numberOfLoops == (i+1))
      {
        int tempC = blockSize - ((numberOfLoops-1)*MAX_THREADS);
        if(tempC < MAX_THREADS)
        { 
          //We can do with less
          threads.x = configNumberOfThreads;
          grid.x = (int)tempC/threads.x;
          if(tempC % threads.x != 0)
            grid.x = grid.x+1;
        }
      }

      //Finally execute the kernel
      forceKernel<<< grid, threads, mem_size >>>(d_xIn,d_vIn,d_ieps2,d_iindex,d_oforce,d_oadoti, d_oNdata);
      CUT_CHECK_ERROR("Kernel execution failed");          
    }
  }
  else //if hu > max_threads
  {  
    //Determine the number of blocks we need
    threads.x = configNumberOfThreads;
    grid.x = (int)blockSize/threads.x;
    if(blockSize % threads.x != 0)
      grid.x = grid.x+1;

    //TODO...
    if(debugPrints)
      printf("Aantal threads: %d grid %d threads %d HU %d\n" , grid.x*threads.x,grid.x, threads.x, blockSize);

    // execute the kernel   
    if(blockSize <= smallKernelEndValue && (particles >=smallKernelDataSetMinSize) )
    {
      //The faster kernel that calculates partial results that get combined
      //on the CPU.

      //Calculate the initial number of extra bundles that we need to start
      int temp1      =  smallKernelBlocks / grid.x;      

      //smallKernelBlocks can be smaller then grid size. Division will
      //give 0 then so, make it 1 incase this happens;
      if(temp1 == 0) temp1 = 1;

      //We can only split the data into a limited number of parts
      //Check if we don't spits it up in more sets then possible
      int parts = tempParticles / usedNumberOfThreads;
      if(temp1 > parts)
        temp1 = parts;

      //Calculate the required number of threads
      localExtraFactor      = temp1;
      grid.x                = localExtraFactor * grid.x;
      localGridSize         = grid.x;
      
      threads.x             = configNumberOfThreads;
      usedNumberOfThreads   = threads.x;
      
      //Copy extra factor to GPU and start execution
      CUDA_SAFE_CALL(cudaMemcpyToSymbol(extraFactor,&localExtraFactor,sizeof(int),0));
      #ifdef TREECODE
        mem_size  = mem_size / 2;
        forceKernelSmallTree<<< grid, threads, mem_size >>>(d_xIn,d_vIn,d_ieps2,d_iindex,d_oforce,d_oadoti, d_oNdata);
      #else
        forceKernelSmall<<< grid, threads, mem_size >>>(d_xIn,d_vIn,d_ieps2,d_iindex,d_oforce,d_oadoti, d_oNdata);
      #endif

      
      CUT_CHECK_ERROR("Kernel execution failed");
    }
    else //Smallkernel
    {
      //The normal kernel
      #ifdef TREECODE
        mem_size  = mem_size / 2;
        forceKernelTree<<< grid, threads, mem_size >>>(d_xIn,d_vIn,d_ieps2,d_iindex,d_oforce,d_oadoti, d_oNdata);
      #else
        forceKernel<<< grid, threads, mem_size >>>(d_xIn,d_vIn,d_ieps2,d_iindex,d_oforce,d_oadoti, d_oNdata);
      #endif

      
      CUT_CHECK_ERROR("Kernel execution failed");
    } //End if > Max threads
    
  }
}

int readBackCUDAResults(int nj, int ni, int index[],double acc[][3], double jerk[][3], double pot[],int* nnindex)
{
  //nj total N
  //ni size of time-step block
 
  // copy result from device to host
  //Memory size depends on number of particles we have calculated
  int bytesToCopy = ni*sizeof(float);

  //If we've used the kernel that splits data then we need to copy
  //more data back from the device.
  if(ni <= smallKernelEndValue && (particles >=smallKernelDataSetMinSize) )
  {
    //Copy the maximum ammount of particles for the smallkernel 
   // bytesToCopy = smallResultsParticles*sizeof(float);  
    bytesToCopy = (grid.x*threads.x)*sizeof(float);  
  }

  CUDA_SAFE_CALL( cudaMemcpy( h_oresadata, d_oforce,bytesToCopy*3, cudaMemcpyDeviceToHost) );
  CUDA_SAFE_CALL( cudaMemcpy( h_oresadotdata, d_oadoti,bytesToCopy*4, cudaMemcpyDeviceToHost) );
  CUDA_SAFE_CALL( cudaMemcpy( h_oNdata, d_oNdata,bytesToCopy*2,  cudaMemcpyDeviceToHost) );

  
 // int ii;

  //Copy results to specified arrays
  if(ni <= smallKernelEndValue && (particles >=smallKernelDataSetMinSize) )
  {
    int tempFactor = localGridSize / localExtraFactor;

    //Clearup temporary results
    for(int i=0; i<ni; i++)
    {   
      acc[i][0]    = 0;
      acc[i][1]    = 0;
      acc[i][2]    = 0;
      jerk[i][0]   = 0;
      jerk[i][1]   = 0;
      jerk[i][2]   = 0;
      pot[i]       = 0;

      //Set nearest neighbour to the first value.
      if(nnindex != NULL)  
      {
        localMinDistanceList[i] = h_oNdata[i].y;
        nnindex[i]              = (int)h_oNdata[i].x;
      }
    }
    
    //Add all partial results
    for(int k=0; k < localExtraFactor; k++)
    {  
      for(int i=0; i<ni; i++)
      {
        acc[i][0]    += h_oresadata[i+tempFactor*k*usedNumberOfThreads].x;
        acc[i][1]    += h_oresadata[i+tempFactor*k*usedNumberOfThreads].y;
        acc[i][2]    += h_oresadata[i+tempFactor*k*usedNumberOfThreads].z;
        jerk[i][0]   += h_oresadotdata[i+tempFactor*k*usedNumberOfThreads].x;
        jerk[i][1]   += h_oresadotdata[i+tempFactor*k*usedNumberOfThreads].y;
        jerk[i][2]   += h_oresadotdata[i+tempFactor*k*usedNumberOfThreads].z;
        pot[i]       += h_oresadotdata[i+tempFactor*k*usedNumberOfThreads].w; 

        //If we save the neighbourhood list we have to find the smallest distance
        //value
        if(nnindex != NULL)
          if(h_oNdata[i+tempFactor*k*usedNumberOfThreads].y  < localMinDistanceList[i])
          {
            localMinDistanceList[i] = h_oNdata[i+tempFactor*k*usedNumberOfThreads].y;
            nnindex[i]              = (int)h_oNdata[i+tempFactor*k*usedNumberOfThreads].x;
          }
          
      }//end particles
    }//end parts
  }
  else
  {  
    for(int i=0; i<ni; i++)
    {   
      acc[i][0]    = h_oresadata[i].x;
      acc[i][1]    = h_oresadata[i].y;
      acc[i][2]    = h_oresadata[i].z;
      jerk[i][0]   = h_oresadotdata[i].x;
      jerk[i][1]   = h_oresadotdata[i].y;
      jerk[i][2]   = h_oresadotdata[i].z;
      pot[i]       = h_oresadotdata[i].w;
      
      //Save neighbourhood list
      if(nnindex != NULL)
      {        
        nnindex[i]   = (int)h_oNdata[i].x;    
      }
    }
  }

  return 0; //todo return 1 als error
}

int calculateNeighbours(double* h2,int blockSize)
{
  //double to float
  for(int i=0; i < blockSize; i++)
  {
    h_h2[i] = (float)h2[i];
  }

   //Copy the h2 values to device
  CUDA_SAFE_CALL( cudaMemcpy( d_h2, h_h2, blockSize*sizeof(float),
                              cudaMemcpyHostToDevice) );

  dim3  grid( 1, 1, 1);
  dim3  threads( 1, 1, 1);

  threads.x = configNumberOfThreads;
  grid.x = (int)blockSize/threads.x;
  if(blockSize % threads.x != 0)
      grid.x = grid.x+1;



  //Shared memory size that we need
  int mem_size =sizeof(float4)*configNumberOfThreads ;

   neighbourKernel<<< grid, threads, mem_size >>>(d_xPred,d_h2, d_ieps2,d_iindex,d_neighbourResults);
   CUT_CHECK_ERROR("Kernel execution failed");

   //Copy the results back in host memory   
  CUDA_SAFE_CALL( cudaMemcpy( h_neighbourResults, d_neighbourResults,sizeof(int)*numberOfPipes*255, cudaMemcpyDeviceToHost) );

  return 1;
}

int getNeighbour(int *pipe, int *max_length,int *nblen, int nbl[])
{
  //printf(" voor h_neigh  pipe %d \n", *pipe );

  int temp = h_neighbourResults[((*pipe))*255];

  if(temp == 0)
  {
    //Overflow
    return 1;
  }
  else if(temp == -1)
  {
    //No neighbours
    *nblen = 0;
    return 0;
  }
  else
  {
    //Copy
    for(int i=1; i <= temp; i++)
    {
      nbl[i-1] = h_neighbourResults[((*pipe)*255)+i];
    }
    *nblen = temp-1;
  }

  return 0;
}



//void startCUDAPredictor(double *ti)
void startCUDAPredictor(double *ti,float *dtj, int *index, float *tj, float *mass, float3 *k18,
                  float3 *j6, float3 *a2, float3 *x, float3 *v, int numberInBuffer)
{
 // printf("predict \n");
  //Copy the timestep to the GPU 
    
  float time = (float)*ti;
  CUDA_SAFE_CALL( cudaMemcpy( d_ipredictTime, &time, sizeof(float),
                              cudaMemcpyHostToDevice) );

  //Start the predictor kernel
  //TODO dit instellen aan hand van N

  int t = 64;
  int g = particles/t;

  dim3 grid(g,1,1);
  dim3 threads(t,1,1);

/*
   CUDA_SAFE_CALL( cudaMemcpy( d_xPred, h_xPred, sizeof(float3)*NMAX,
                              cudaMemcpyHostToDevice) );
   CUDA_SAFE_CALL( cudaMemcpy( d_vPred, h_vPred, sizeof(float3)*NMAX,
                              cudaMemcpyHostToDevice) );

   free(h_vPred);
   free(h_xPred);*/
 //predictKernel<<< grid, threads, 0 >>>(
  //  d_ipredictTime, d_ix, d_iv, d_itj, NULL, d_xPred, d_vPred, d_oforce, d_oadoti);

 // printf("Tijd in kernel %f \n", cutGetTimerValue(start));

 CUT_CHECK_ERROR("Kernel execution failed");

}


/* This functions free's up all allocated memory on the device */
void cleanUpCUDA()
{ 
  printf("Kirin execution summary \n");
  printf("Total block sum: %d  \t total number of calls: %d \t average time-block size: %f \n", blockSizeSum,numberOfCalls, blockSizeSum/(double)numberOfCalls);

  free(h_ixdata);
  free(h_ivdata);
  
  free(h_oresadata);
  free(h_oresadotdata);
  free(h_oNdata);
  free(localMinDistanceList);
  free(h_h2); 
  free(h_neighbourResults);
  free(h_xIn);
  free(h_vIn);

 CUDA_SAFE_CALL(cudaFree(d_neighbourResults));
 CUDA_SAFE_CALL(cudaFree(d_ix));
 CUDA_SAFE_CALL(cudaFree(d_iv));
 CUDA_SAFE_CALL(cudaFree(d_h2));
 
 CUDA_SAFE_CALL(cudaFree(d_iindex));
 CUDA_SAFE_CALL(cudaFree(d_ihighest));
 CUDA_SAFE_CALL(cudaFree(d_ipredictTime));

 CUDA_SAFE_CALL(cudaFree(d_oforce));
 CUDA_SAFE_CALL(cudaFree(d_oadoti));
 CUDA_SAFE_CALL(cudaFree(d_oNdata));
 CUDA_SAFE_CALL(cudaFree(d_xPred));
 CUDA_SAFE_CALL(cudaFree(d_vPred));

 CUDA_SAFE_CALL(cudaFree(d_xIn));
 CUDA_SAFE_CALL(cudaFree(d_vIn));

}
