#ifndef _TEMPLATE_KERNEL_H_
#define _TEMPLATE_KERNEL_H_

#include <stdio.h>

//Textures 
texture<float4, 1, cudaReadModeElementType> texX;
texture<float4, 1, cudaReadModeElementType> texV;

__global__ void
forceKernel(float4* xIn, float3* vIn,real *eps2, int* index, float3* g_odata, float4* g_adoti, float2* g_oNdata)
{
  float3 forceNew = make_float3(0,0,0) ;
  float4 adotiNew = make_float4(0,0,0,0);

  extern __shared__ float4 sxData[];

  //The number of the current particle that we do calculations on
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //Increase outputIndex incase we work around the 5second limit
  outputIndex += indexModifier;
  
  //Define variables
  float   otherMass;
  float3  dx;
  float3  dv;
  float   r2;
  float   xdotv;
  float   r2inv;
  float   rinv ;
  float   r3inv;
  float   r5inv;
  float   xdotvr5inv;
  float3  tempr3InvDv;
  float3  tempxdotVr5Inv;

  //Each thread stands for an output and an associated particle 
  float4 xOfInterest = xIn[outputIndex];
  float3 vOfInterest = vIn[outputIndex];


  float distance;
  float minDistance = 99999;
  int   indexMinDistance = 0;
  
 
  //Aantal stappen moet gelijk aan totaal N / aantal threads
 for(int k=0; k < numberOfParticles / blockDim.x; k++)
  {    
    //Read data of current step to shared memory
    sxData[threadIdx.x*2] = tex1Dfetch(texX,(int)( threadIdx.x + k*blockDim.x));   
    sxData[threadIdx.x*2+1] = tex1Dfetch(texV,(int)( threadIdx.x + k*blockDim.x));
   
     __syncthreads();

    float4 tempXData;
    float4 tempVData;

    //For each data element in the shared memory buffer => each thread   
    for(int i=0; i < blockDim.x; i++)
    {
      //Calculate the force between i and outputIndex
      tempXData = sxData[i*2];     
      tempVData = sxData[i*2+1];

      //Mass is 4th component of the x data
      otherMass = tempXData.w;
     
      dx = make_float3(tempXData.x - xOfInterest.x, tempXData.y - xOfInterest.y, tempXData.z - xOfInterest.z);
      dv = make_float3(tempVData.x - vOfInterest.x, tempVData.y - vOfInterest.y, tempVData.z - vOfInterest.z);  

      //See if this star is closer then any other stars
      distance  = dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      if(distance*otherMass < minDistance && distance*otherMass > 0)
      {
        minDistance       = distance*otherMass;
        indexMinDistance  = i+k*blockDim.x;
      }

      r2 = *eps2+distance;
     //  r2            = *eps2+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      //TODOr2              = constantEps+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;      
      xdotv           = dx.x*dv.x + dx.y*dv.y + dx.z*dv.z;
      r2inv           = 1.0/r2;

      if((*eps2 == 0) && (distance == 0))
      {
        r2inv = 1;
        rinv = 0;
      }
      else
      {
        rinv            = sqrt(r2inv);
      }

      r3inv           = r2inv*rinv;
      r5inv           = r2inv*r3inv;
      xdotvr5inv      = 3.0*xdotv*r5inv;
          
      tempr3InvDv     = make_float3(r3inv*dv.x,r3inv*dv.y,r3inv*dv.z);
      tempxdotVr5Inv  = make_float3(xdotvr5inv * dx.x, xdotvr5inv * dx.y, xdotvr5inv * dx.z);
      
      forceNew.x = forceNew.x + otherMass*r3inv*dx.x;
      forceNew.y = forceNew.y + otherMass*r3inv*dx.y;
      forceNew.z = forceNew.z + otherMass*r3inv*dx.z;

      adotiNew.x = adotiNew.x + otherMass*(tempr3InvDv.x - tempxdotVr5Inv.x);
      adotiNew.y = adotiNew.y + otherMass*(tempr3InvDv.y - tempxdotVr5Inv.y);
      adotiNew.z = adotiNew.z + otherMass*(tempr3InvDv.z - tempxdotVr5Inv.z);
      adotiNew.w = adotiNew.w - otherMass*rinv;
      
    }// end for

    //Go on when all threads have calculated their part
    __syncthreads();
  }
    
  //Remove the self interaction
  if(*eps2 != 0)
   adotiNew.w +=  sqrt(1.0/ *eps2)*xOfInterest.w;
 
  //Get the original output index (before the index[] lookup  
  outputIndex   = blockIdx.x * blockDim.x + threadIdx.x;
  outputIndex   += indexModifier;

  //Save the results
  g_odata[outputIndex]    =  forceNew;
  g_adoti[outputIndex]    =  adotiNew;
  g_oNdata[outputIndex].x =  indexMinDistance;
}

//Alternative version of the kernel above, this one does partial calculations
//that get combined later on on the PC.
__global__ void
forceKernelSmall(float4* xIn, float3* vIn, real *eps2,  int* index, float3* g_odata, float4* g_adoti, float2* g_oNdata)
{
  float3 forceNew = make_float3(0,0,0) ;
  float4 adotiNew = make_float4(0,0,0,0);

  extern __shared__ float4 sxData[];

  //The unique thread number
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //To determine which particle we have to processes we have to find out in 
  //which block we are in relation to the other blocks.
  int temp  = (gridDim.x / extraFactor);
  temp      = blockIdx.x % temp;
  //Temp now contains the block we are, so determine the output index
  outputIndex = temp*blockDim.x  + threadIdx.x;

  //No orderd data so need to look up the index
  //outputIndex = index[outputIndex];

  //Increase outputIndex with the value of indexModifier
  //This is for the 5 seconds workaround
 //TODO outputIndex += indexModifier;
  
  //Define variables
  float   otherMass;
  float3  dx;
  float3  dv;
  float   r2;
  float   xdotv;
  float   r2inv;
  float   rinv ;
  float   r3inv;
  float   r5inv;
  float   xdotvr5inv;
  float3  tempr3InvDv;
  float3  tempxdotVr5Inv;


  //Each thread stands for an output and an associated particle
  //Save that particles data in a local variable
//  float4 xOfInterest = tex1Dfetch(texX,(int)( outputIndex));
//  float4 vOfInterest = tex1Dfetch(texV,(int)( outputIndex));
  float4 xOfInterest = xIn[outputIndex];
  float3 vOfInterest = vIn[outputIndex];


  float distance;
  float minDistance = 99999;
  int   indexMinDistance = 0;

  //We have to find out how many parts we have to process and at what part
  //we have to start. First we determine the total number of parts. 
  //Then we calculate the new number of parts that each bundle has to process.
  int oldNumberofParts = numberOfParticles / blockDim.x;
  int newNumberofParts = oldNumberofParts / extraFactor;
 
  //Offset is determined by finding out how many blocks there are before us
  //(eg which block are we), this is multiplied by the number of particles per block.
  int offset  = blockIdx.x/(gridDim.x / extraFactor)*(blockDim.x*newNumberofParts);

  //Since we can not always spread the total number of parts evenly over the number
  //of blocks, we will add the parts that are left over to the last blocks.
  //First we determine if we need to increase the number of parts of that it is evenly
  //distributed.
  if((blockDim.x * (extraFactor*newNumberofParts))  <  numberOfParticles)
  {
    //We have to process aditional blocks

    //De laatste extra factor blokken moeten een aantal parts extra doen
    //om alle particles te verwerken.

    //Determine if we are the last block for this particle.
    if((blockIdx.x/(gridDim.x / extraFactor)) == (extraFactor-1))
    {
      //We are the last so increase the total number of parts that we have to process
      newNumberofParts += (numberOfParticles-(blockDim.x * extraFactor * newNumberofParts))/blockDim.x;
    }
  }

  //From here on everything works as in the other kernel
   
  for(int k=0; k < newNumberofParts; k++)
  {    
    //Read data of current step to shared memory
    sxData[threadIdx.x*2] = tex1Dfetch(texX,(int)( threadIdx.x + offset+k*blockDim.x));
    sxData[threadIdx.x*2+1] = tex1Dfetch(texV,(int)( threadIdx.x + offset+k*blockDim.x));
    

    __syncthreads();
     
    float4 tempXData;
    float4 tempVData;


    //For each data element in the shared memory buffer => each thread   
    for(int i=0; i < blockDim.x; i++)
    {
      //Calculate the force between i and outputIndex      
      tempXData = sxData[i*2];
      tempVData = sxData[i*2+1];

      //Mass is 4th component of the x data
      otherMass = tempXData.w;

      dx = make_float3(tempXData.x - xOfInterest.x, tempXData.y - xOfInterest.y, tempXData.z - xOfInterest.z);
      dv = make_float3(tempVData.x - vOfInterest.x, tempVData.y - vOfInterest.y, tempVData.z - vOfInterest.z);  
  
      distance  = dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;

      //See if this star is closer then any other star
      if(distance*otherMass < minDistance && distance*otherMass > 0)
      {
        minDistance       = distance*otherMass;
        indexMinDistance  = offset+i+k*blockDim.x;
      }

      //r2            = *eps2+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      r2              = *eps2+distance;      
      xdotv           = dx.x*dv.x + dx.y*dv.y + dx.z*dv.z;
      r2inv           = 1.0/r2;    
     
      if((*eps2 == 0) && (distance == 0))
      {
        r2inv = 1;
        rinv = 0;
      }
      else
      {
        rinv            = sqrt(r2inv);
      }
      //r2inv           = __fdividef(1.0,r2);    
     // rinv            = sqrt(r2inv);
      r3inv           = r2inv*rinv;
      r5inv           = r2inv*r3inv;
      xdotvr5inv      = 3.0*xdotv*r5inv;
          
      tempr3InvDv     = make_float3(r3inv*dv.x,r3inv*dv.y,r3inv*dv.z);
      tempxdotVr5Inv  = make_float3(xdotvr5inv * dx.x, xdotvr5inv * dx.y, xdotvr5inv * dx.z);
      
      forceNew.x = forceNew.x + otherMass*r3inv*dx.x;
      forceNew.y = forceNew.y + otherMass*r3inv*dx.y;
      forceNew.z = forceNew.z + otherMass*r3inv*dx.z;

      adotiNew.x = adotiNew.x + otherMass*(tempr3InvDv.x - tempxdotVr5Inv.x);
      adotiNew.y = adotiNew.y + otherMass*(tempr3InvDv.y - tempxdotVr5Inv.y);
      adotiNew.z = adotiNew.z + otherMass*(tempr3InvDv.z - tempxdotVr5Inv.z);
      adotiNew.w = adotiNew.w - otherMass*rinv;
    }// end for
    
    //Go on when all threads have calculated their part
    __syncthreads();
  }
    
  //Remove the self interaction
  //But only in 1 part, we do it in the first part/block that processes this particle
  //Remove the self interaction
  if(*eps2 != 0)
    if(blockIdx.x < (gridDim.x / extraFactor))
      adotiNew.w +=  sqrt(1.0/ *eps2)*xOfInterest.w;
 
  //Get the original output location
  outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //Save the results
  g_odata[outputIndex] =  forceNew;
  g_adoti[outputIndex] =  adotiNew;
  g_oNdata[outputIndex] =  make_float2(indexMinDistance, minDistance);
}


__global__ void
neighbourKernel(float4* x, float* h2,real *eps2, int* index, int * neighbourList)
{
   //The number of the current particle that we do calculations on
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

    //crash als outputIndex 19136 is...
  float h2OfInterest = h2[outputIndex];
  //Extra lookup that is needed since data is not ordered in memory
  outputIndex = index[outputIndex];

  extern __shared__ float4 sxData[];
  //Determine the neighbours for the various particles

   //Each thread stands for an output and an associated particle 
  float4 xOfInterest = tex1Dfetch(texX,(int)( outputIndex));
  //float4 vOfInterest = texfetch(texV,(int)( outputIndex));

  float distance;
  int numberOfNeighbours = 0;


 
  //Aantal stappen moet gelijk aan totaal N / aantal threads
 for(int k=0; k < numberOfParticles / blockDim.x; k++)
  {    
    //Read data of current step to shared memory
    sxData[threadIdx.x] = tex1Dfetch(texX,(int)( threadIdx.x + k*blockDim.x));   
   
    __syncthreads();

    float4 tempXData;

    //For each data element in the shared memory buffer => each thread   
    for(int i=0; i < blockDim.x; i++)
    {
      //Calculate the force between i and outputIndex
      tempXData = sxData[i];     
     
      float3 dx = make_float3(tempXData.x - xOfInterest.x, tempXData.y - xOfInterest.y, tempXData.z - xOfInterest.z);

      //See if this star is closer then any other stars
      distance  = dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      if(distance < h2OfInterest && (numberOfNeighbours < 254) && (i+k*blockDim.x != outputIndex))
      {
        numberOfNeighbours++;
        neighbourList[(( blockIdx.x * blockDim.x + threadIdx.x)*255) + numberOfNeighbours] = i+k*blockDim.x;
        //was <= 254
        //neighbours[numberOfNeighbours] = i+k*blockDim.x;
        //numberOfNeighbours++;
      }
    }// end for

    //Go on when all threads have calculated their part
    __syncthreads();
  }
    
  
  //Get the original output index (before the index[] lookup  
  outputIndex   = blockIdx.x * blockDim.x + threadIdx.x;
  
  //Now add some offset for the output
 // float tempO = outputIndex*255.0;
  outputIndex = outputIndex*255;


  if(numberOfNeighbours == 254)
  {
    //Overflow, only set first neighbour and put a 0 in it.
    neighbourList[outputIndex] = 0;
  }
  else if(numberOfNeighbours == 0)
  {
    //No neighbours, put an ehmnm -1
    neighbourList[outputIndex] = -1;
  }
  else
  {
    //Set the number of found neighbours
    neighbourList[outputIndex] = numberOfNeighbours;
    //Save the neighbours/ in global memory
   // for(int i=1; i < numberOfNeighbours; i++)
     //  neighbourList[outputIndex+i] = neighbours[i];

  }

 //neighbourList[outputIndex] = 123456;

}







__global__ void
forceKernelTree(float4* xIn, float3* vIn,real *eps2, int* index, float3* g_odata, float4* g_adoti, float2* g_oNdata)
{
  float3 forceNew = make_float3(0,0,0) ;
  float4 adotiNew = make_float4(0,0,0,0);

  extern __shared__ float4 sxData[];

  //The number of the current particle that we do calculations on
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //Increase outputIndex incase we work around the 5second limit
  outputIndex += indexModifier;
  
  //Define variables
  float   otherMass;
  float3  dx;
  float   r2;
  float   r2inv;
  float   rinv ;
  float   r3inv;


  //Extra lookup that is needed since data is not ordered in memory
 // outputIndex = index[outputIndex];
  //Each thread stands for an output and an associated particle 
 // float4 xOfInterest = tex1Dfetch(texX,(int)( outputIndex));
 // float4 vOfInterest = tex1Dfetch(texV,(int)( outputIndex));
  float4 xOfInterest = xIn[outputIndex];
 
  float distance;
 
  //Aantal stappen moet gelijk aan totaal N / aantal threads
 for(int k=0; k < numberOfParticles / blockDim.x; k++)
  {    
    //Read data of current step to shared memory
    sxData[threadIdx.x] = tex1Dfetch(texX,(int)( threadIdx.x + k*blockDim.x));   
   
     __syncthreads();

    float4 tempXData;   

    //For each data element in the shared memory buffer => each thread   
    for(int i=0; i < blockDim.x; i++)
    {
      //Calculate the force between i and outputIndex
      tempXData = sxData[i];     
  
      //Mass is 4th component of the x data
      otherMass = tempXData.w;
     
      dx = make_float3(tempXData.x - xOfInterest.x, tempXData.y - xOfInterest.y, tempXData.z - xOfInterest.z);
   
      distance = dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      r2 = *eps2+distance;
     //  r2            = *eps2+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      //TODOr2              = constantEps+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;  
      r2inv           = 1.0/r2;

      if((*eps2 == 0) && (distance == 0))
      {
        r2inv = 1;
        rinv = 0;
      }
      else
      {
        rinv            = sqrt(r2inv);
      }

      r3inv           = r2inv*rinv;
          
      forceNew.x = forceNew.x + otherMass*r3inv*dx.x;
      forceNew.y = forceNew.y + otherMass*r3inv*dx.y;
      forceNew.z = forceNew.z + otherMass*r3inv*dx.z;

      adotiNew.w = adotiNew.w - otherMass*rinv;      
    }// end for

    //Go on when all threads have calculated their part
    __syncthreads();
  }
    
  //Remove the self interaction
  if(*eps2 != 0)
   adotiNew.w +=  sqrt(1.0/ *eps2)*xOfInterest.w;
 
  //Get the original output index (before the index[] lookup  
  outputIndex   = blockIdx.x * blockDim.x + threadIdx.x;
  outputIndex   += indexModifier;

  //Save the results
  g_odata[outputIndex]    =  forceNew;
  g_adoti[outputIndex]    =  adotiNew; 
}







//Alternative version of the kernel above, this one does partial calculations
//that get combined later on on the PC.
__global__ void
forceKernelSmallTree(float4* xIn, float3* vIn, real *eps2,  int* index, float3* g_odata, float4* g_adoti, float2* g_oNdata)
{
  float3 forceNew = make_float3(0,0,0) ;
  float4 adotiNew = make_float4(0,0,0,0);

  extern __shared__ float4 sxData[];

  //The unique thread number
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //To determine which particle we have to processes we have to find out in 
  //which block we are in relation to the other blocks.
  int temp  = (gridDim.x / extraFactor);
  temp      = blockIdx.x % temp;
  //Temp now contains the block we are, so determine the output index
  outputIndex = temp*blockDim.x  + threadIdx.x;

  //No orderd data so need to look up the index
  //outputIndex = index[outputIndex];

  //Increase outputIndex with the value of indexModifier
  //This is for the 5 seconds workaround
 //TODO outputIndex += indexModifier;
  
  //Define variables
  float   otherMass;
  float3  dx;
  float   r2; 
  float   r2inv;
  float   rinv ;
  float   r3inv;
 

  //Each thread stands for an output and an associated particle
  //Save that particles data in a local variable
//  float4 xOfInterest = tex1Dfetch(texX,(int)( outputIndex));
//  float4 vOfInterest = tex1Dfetch(texV,(int)( outputIndex));
  float4 xOfInterest = xIn[outputIndex];

  float distance;
 
  //We have to find out how many parts we have to process and at what part
  //we have to start. First we determine the total number of parts. 
  //Then we calculate the new number of parts that each bundle has to process.
  int oldNumberofParts = numberOfParticles / blockDim.x;
  int newNumberofParts = oldNumberofParts / extraFactor;
 
  //Offset is determined by finding out how many blocks there are before us
  //(eg which block are we), this is multiplied by the number of particles per block.
  int offset  = blockIdx.x/(gridDim.x / extraFactor)*(blockDim.x*newNumberofParts);

  //Since we can not always spread the total number of parts evenly over the number
  //of blocks, we will add the parts that are left over to the last blocks.
  //First we determine if we need to increase the number of parts of that it is evenly
  //distributed.
  if((blockDim.x * (extraFactor*newNumberofParts))  <  numberOfParticles)
  {
    //We have to process aditional blocks

    //De laatste extra factor blokken moeten een aantal parts extra doen
    //om alle particles te verwerken.

    //Determine if we are the last block for this particle.
    if((blockIdx.x/(gridDim.x / extraFactor)) == (extraFactor-1))
    {
      //We are the last so increase the total number of parts that we have to process
      newNumberofParts += (numberOfParticles-(blockDim.x * extraFactor * newNumberofParts))/blockDim.x;
    }
  }

  //From here on everything works as in the other kernel
   
  for(int k=0; k < newNumberofParts; k++)
  {    
    //Read data of current step to shared memory 
    sxData[threadIdx.x] = tex1Dfetch(texX,(int)( threadIdx.x + offset+k*blockDim.x));

    __syncthreads();
     
    float4 tempXData;


    //For each data element in the shared memory buffer => each thread   
    for(int i=0; i < blockDim.x; i++)
    {
      //Calculate the force between i and outputIndex      
      tempXData = sxData[i];
 
      //Mass is 4th component of the x data
      otherMass = tempXData.w;

      dx = make_float3(tempXData.x - xOfInterest.x, tempXData.y - xOfInterest.y, tempXData.z - xOfInterest.z);
     
      distance  = dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;

      //r2            = *eps2+dx.x*dx.x + dx.y*dx.y + dx.z*dx.z;
      r2              = *eps2+distance;         
      r2inv           = 1.0/r2;    
     
      if((*eps2 == 0) && (distance == 0))
      {
        r2inv = 1;
        rinv = 0;
      }
      else
      {
        rinv            = sqrt(r2inv);
      }
      //r2inv           = __fdividef(1.0,r2);    
     // rinv            = sqrt(r2inv);
      r3inv           = r2inv*rinv;
      
      forceNew.x = forceNew.x + otherMass*r3inv*dx.x;
      forceNew.y = forceNew.y + otherMass*r3inv*dx.y;
      forceNew.z = forceNew.z + otherMass*r3inv*dx.z;

      adotiNew.w = adotiNew.w - otherMass*rinv;
    }// end for
    
    //Go on when all threads have calculated their part
    __syncthreads();
  }
    
  //Remove the self interaction
  //But only in 1 part, we do it in the first part/block that processes this particle
  //Remove the self interaction
  if(*eps2 != 0)
    if(blockIdx.x < (gridDim.x / extraFactor))
      adotiNew.w +=  sqrt(1.0/ *eps2)*xOfInterest.w;
 
  //Get the original output location
  outputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  //Save the results
  g_odata[outputIndex] =  forceNew;
  g_adoti[outputIndex] =  adotiNew;
}





/*


__global__ void
predictKernel(real *time, float3* x, float3* v, real* ti, int *indexModifier, float3* xPred, float3* vPred, float3* force, float4* adoti)
{
  unsigned int outputIndex = blockIdx.x * blockDim.x + threadIdx.x;
  real dt2half, dt3over6, dt;
  
  //Calculate the time values
  dt        = *time - ti[outputIndex];
  dt2half   = 0.5*dt*dt;
  dt3over6  = 1.0/3.0*dt*dt2half;

  float3 localX     = x[outputIndex];
  float3 localV     = v[outputIndex];
  float3 localForce = force[outputIndex];
  float4 localJerk  = adoti[outputIndex];

  float3 localXPred, localVPred;

  //todo dit maken met minder geheugen aanroepen...
  
  //Calculate the predictions
  localXPred.x = localX.x + dt*localV.x + dt2half*localForce.x  + dt3over6*localJerk.x;
  localXPred.y = localX.y + dt*localV.y + dt2half*localForce.y  + dt3over6*localJerk.y;
  localXPred.z = localX.z + dt*localV.z + dt2half*localForce.z  + dt3over6*localJerk.z;

  localVPred.x = localV.x + dt*localForce.x + dt2half*localJerk.x;
  localVPred.y = localV.y + dt*localForce.y + dt2half*localJerk.y;
  localVPred.z = localV.z + dt*localForce.z + dt2half*localJerk.z;

  xPred[outputIndex] = localXPred;
  vPred[outputIndex] = localVPred;

}*/
#endif // #ifndef _TEMPLATE_KERNEL_H_
